# Copyright (C) 2007-2021, Stefan Schwarzer <sschwarzer@sschwarzer.net>
# and ftputil contributors (see `doc/contributors.txt`)
# See the file LICENSE for licensing terms.

import io
import ntpath
import os
import shutil

import ftputil
import ftputil.path_encoding
import ftputil.sync


# Assume the test subdirectories are or will be in the current directory.
TEST_ROOT = os.getcwd()


class TestLocalToLocal:
    def setup_method(self, method):
        if not os.path.exists("test_empty"):
            os.mkdir("test_empty")
        if os.path.exists("test_target"):
            shutil.rmtree("test_target")
        os.mkdir("test_target")

    def test_sync_empty_dir(self):
        source = ftputil.sync.LocalHost()
        target = ftputil.sync.LocalHost()
        syncer = ftputil.sync.Syncer(source, target)
        source_dir = os.path.join(TEST_ROOT, "test_empty")
        target_dir = os.path.join(TEST_ROOT, "test_target")
        syncer.sync(source_dir, target_dir)

    def test_source_with_and_target_without_slash(self):
        source = ftputil.sync.LocalHost()
        target = ftputil.sync.LocalHost()
        syncer = ftputil.sync.Syncer(source, target)
        source_dir = os.path.join(TEST_ROOT, "test_source/")
        target_dir = os.path.join(TEST_ROOT, "test_target")
        syncer.sync(source_dir, target_dir)


# Helper classes for `TestUploadFromWindows`


class LocalWindowsHost(ftputil.sync.LocalHost):
    def __init__(self):
        self.path = ntpath
        self.sep = "\\"

    def open(self, path, mode):
        # Just return a dummy file object.
        return io.StringIO("")

    def walk(self, root):
        """
        Return a list of tuples as `os.walk`, but use tuples as if the
        directory structure was

        <root>
            dir1
                dir11
                file1
                file2

        where <root> is the string passed in as `root`.
        """
        join = ntpath.join
        return [
            (root, [join(root, "dir1")], []),
            (join(root, "dir1"), ["dir11"], ["file1", "file2"]),
        ]


class DummyFTPSession:
    def pwd(self):
        return "/"


class DummyFTPPath:
    def abspath(self, path):
        # Don't care here if the path is absolute or not.
        return path

    def isdir(self, path):
        return "dir" in path

    def isfile(self, path):
        return "file" in path


class ArgumentCheckingFTPHost(ftputil.FTPHost):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.path = DummyFTPPath()

    def _make_session(self, *args, **kwargs):
        self._encoding = ftputil.path_encoding.DEFAULT_ENCODING
        return DummyFTPSession()

    def mkdir(self, path):
        assert "\\" not in path

    def open(self, path, mode):
        assert "\\" not in path
        return io.StringIO("")


class TestUploadFromWindows:
    def test_no_mixed_separators(self):
        source = LocalWindowsHost()
        target = ArgumentCheckingFTPHost()
        local_root = ntpath.join("some", "directory")
        syncer = ftputil.sync.Syncer(source, target)
        # If the following call raises any `AssertionError`s, the test
        # framework will catch them and show them.
        syncer.sync(local_root, "not_used_by_ArgumentCheckingFTPHost")



# Copyright (C) 2003-2021, Stefan Schwarzer <sschwarzer@sschwarzer.net>
# and ftputil contributors (see `doc/contributors.txt`)
# See the file LICENSE for licensing terms.

"""
ftputil.path - simulate `os.path` for FTP servers
"""

import os
import posixpath
import stat

import ftputil.error
import ftputil.tool


# The `_Path` class shouldn't be used directly by clients of the ftputil
# library.
__all__ = []


class _Path:
    """
    Support class resembling `os.path`, accessible from the `FTPHost` object,
    e. g. as `FTPHost().path.abspath(path)`.

    Hint: substitute `os` with the `FTPHost` object.
    """

    # `_Path` needs to provide all methods of `os.path`.
    # pylint: disable=too-many-instance-attributes

    def __init__(self, host):
        self._host = host
        self._encoding = host._encoding
        # Delegate these methods to the `posixpath` module because they don't
        # need file system access but work on the path strings (possibly
        # extracted from `PathLike` objects).
        # pylint: disable=invalid-name
        pp = posixpath
        self.basename = pp.basename
        self.commonprefix = pp.commonprefix
        self.dirname = pp.dirname
        self.isabs = pp.isabs
        self.join = pp.join
        self.normcase = pp.normcase
        self.normpath = pp.normpath
        self.split = pp.split
        self.splitdrive = pp.splitdrive
        self.splitext = pp.splitext

    def abspath(self, path):
        """
        Return an absolute path.
        """
        # Don't use `raise_for_empty_path` here since Python itself doesn't
        # raise an exception and just returns the current directory.
        original_path = path
        path = ftputil.tool.as_str_path(path, encoding=self._encoding)
        if not self.isabs(path):
            path = self.join(self._host.getcwd(), path)
        return ftputil.tool.same_string_type_as(
            os.fspath(original_path), self.normpath(path), self._encoding
        )

    def exists(self, path):
        """
        Return true if the path exists.
        """
        if path in ["", b""]:
            return False
        try:
            lstat_result = self._host.lstat(path, _exception_for_missing_path=False)
            return lstat_result is not None
        except ftputil.error.RootDirError:
            return True

    def getmtime(self, path):
        """
        Return the timestamp for the last modification for `path` as a float.

        This will raise `PermanentError` if the path doesn't exist, but maybe
        other exceptions depending on the state of the server (e. g. timeout).
        """
        ftputil.tool.raise_for_empty_path(path)
        return self._host.stat(path).st_mtime

    def getsize(self, path):
        """
        Return the size of the `path` item as an integer.

        This will raise `PermanentError` if the path doesn't exist, but maybe
        raise other exceptions depending on the state of the server (e. g.
        timeout).
        """
        ftputil.tool.raise_for_empty_path(path)
        return self._host.stat(path).st_size

    # Check whether a path is a regular file/dir/link. For the first two cases
    # follow links (like in `os.path`).
    #
    # Implementation note: The previous implementations simply called `stat` or
    # `lstat` and returned `False` if they ended with raising a
    # `PermanentError`. That exception usually used to signal a missing path.
    # This approach has the problem, however, that exceptions caused by code
    # earlier in `lstat` are obscured by the exception handling in `isfile`,
    # `isdir` and `islink`.

    def _is_file_system_entity(self, path, dir_or_file):
        """
        Return `True` if `path` represents the file system entity described by
        `dir_or_file` ("dir" or "file").

        Return `False` if `path` isn't a directory or file, respectively or if
        `path` leads to an infinite chain of links.
        """
        assert dir_or_file in ["dir", "file"]
        # Consider differences between directories and files.
        if dir_or_file == "dir":
            should_look_for_dir = True
            stat_function = stat.S_ISDIR
        else:
            should_look_for_dir = False
            stat_function = stat.S_ISREG
        #
        path = ftputil.tool.as_str_path(path, encoding=self._encoding)
        #  Workaround if we can't go up from the current directory. The result
        #  from `getcwd` should already be normalized.
        if self.normpath(path) == self._host.getcwd():
            return should_look_for_dir
        try:
            stat_result = self._host.stat(path, _exception_for_missing_path=False)
        except ftputil.error.RecursiveLinksError:
            return False
        except ftputil.error.RootDirError:
            return should_look_for_dir
        else:
            if stat_result is None:
                # Non-existent path
                return False
            else:
                return stat_function(stat_result.st_mode)

    def isdir(self, path):
        """
        Return true if the `path` exists and corresponds to a directory (no
        link).

        A non-existing path does _not_ cause a `PermanentError`, instead return
        `False`.
        """
        if path in ["", b""]:
            return False
        return self._is_file_system_entity(path, "dir")

    def isfile(self, path):
        """
        Return true if the `path` exists and corresponds to a regular file (no
        link).

        A non-existing path does _not_ cause a `PermanentError`, instead return
        `False`.
        """
        if path in ["", b""]:
            return False
        return self._is_file_system_entity(path, "file")

    def islink(self, path):
        """
        Return true if the `path` exists and is a link.

        A non-existing path does _not_ cause a `PermanentError`, instead return
        `False`.
        """
        path = ftputil.tool.as_str_path(path, encoding=self._encoding)
        if path == "":
            return False
        try:
            lstat_result = self._host.lstat(path, _exception_for_missing_path=False)
        except ftputil.error.RootDirError:
            return False
        else:
            if lstat_result is None:
                # Non-existent path
                return False
            else:
                return stat.S_ISLNK(lstat_result.st_mode)

    def walk(self, top, func, arg):
        """
        Directory tree walk with callback function.

        For each directory in the directory tree rooted at top
        (including top itself, but excluding "." and ".."), call
        func(arg, dirname, fnames). dirname is the name of the
        directory, and fnames a list of the names of the files and
        subdirectories in dirname (excluding "." and "..").  func may
        modify the fnames list in-place (e.g. via del or slice
        assignment), and walk will only recurse into the
        subdirectories whose names remain in fnames; this can be used
        to implement a filter, or to impose a specific order of
        visiting.  No semantics are defined for, or required of, arg,
        beyond that arg is always passed to func.  It can be used,
        e.g., to pass a filename pattern, or a mutable object designed
        to accumulate statistics.  Passing None for arg is common.
        """
        ftputil.tool.raise_for_empty_path(top, path_argument_name="top")
        top = ftputil.tool.as_str_path(top, encoding=self._encoding)
        # This code (and the above documentation) is taken from `posixpath.py`,
        # with slight modifications.
        try:
            names = self._host.listdir(top)
        except OSError:
            return
        func(arg, top, names)
        for name in names:
            name = self.join(top, name)
            try:
                stat_result = self._host.lstat(name)
            except OSError:
                continue
            if stat.S_ISDIR(stat_result[stat.ST_MODE]):
                self.walk(name, func, arg)



# Copyright (C) 2003-2019, Stefan Schwarzer <sschwarzer@sschwarzer.net>
# and ftputil contributors (see `doc/contributors.txt`)
# See the file LICENSE for licensing terms.

import io

import ftputil


# Since `io.BytesIO` and `io.StringIO` are built-in, they can't be patched with
# `unittest.mock.patch`. However, derived classes can be mocked. Mocking is
# useful to test the arguments of `write` calls, i. e. whether the expected
# data was written.
class MockableBytesIO(io.BytesIO):
    pass


class MockableStringIO(io.StringIO):
    pass


# Factory to produce `FTPHost`-like classes from a given `FTPHost` class and
# (usually) a given `MockSession` class.
def ftp_host_factory(session_factory, ftp_host_class=ftputil.FTPHost):
    return ftp_host_class(
        "dummy_host", "dummy_user", "dummy_password", session_factory=session_factory
    )


def dir_line(
    mode_string="-r--r--r--",
    nlink=1,
    user="dummy_user",
    group="dummy_group",
    size=512,
    date_=None,
    datetime_=None,
    name="dummy_name",
    link_target=None,
):
    """
    Return a line as it would be returned by an FTP `DIR` invocation.

    Some values are handled specially:

    - One of `date_` or `datetime_` must be given, the other must be `None`.

      If `date_` is given, it must be a `datetime.date` object. The timestamp
      in the `DIR` line is formatted like "Apr 22 2019", with the concrete
      value taken from the `date_` object.

      If `datetime_` is given, it must be a `datetime.datetime` object. The
      timestamp in the `DIR` line is formatted like "Apr 22 16:50", with the
      concrete value taken from the `datetime_` object. Timezone information in
      the `datetime_` object is ignored.

    - If `link_target` is left at the default `None`, the name part is the
      value from the `name` argument. If `link_target` isn't `None`, the name
      part of the `DIR` line is formatted as "name -> link_target".

    Note that the spacing between the parts of the line isn't necessarily
    exactly what you'd get from an FTP server because the parser in ftputil
    doesn't take the exact amount of spaces into account, so the `DIR` lines
    don't have to be that accurate.

    Examples:

      # Result:
      # "drwxr-xr-x  2  dummy_user dummy_group  182  Apr 22 16:50  file_name"
      line = dir_line(mode_string="-rw-rw-r--",
                      nlink=2,
                      size=182,
                      datetime=datetime.datetime.now(),
                      name="file_name")

      # Result:
      # "drwxr-xr-x  1  dummy_user dummy_group  512  Apr 22 2019  dir_name -> dir_target"
      line = dir_line(mode_string="drwxr-xr-x",
                      date=datetime.date.today(),
                      name="dir_name",
                      link_target="dir_target")
    """
    # Date or datetime. We must have exactly one of `date_` and `datetime_`
    # set. The other value must be `None`.
    assert [date_, datetime_].count(
        None
    ) == 1, "specify exactly one of `date_` and `datetime_`"
    if date_:
        datetime_string = date_.strftime("%b %d %Y")
    else:
        datetime_string = datetime_.strftime("%b %d %H:%M")
    # Name, possibly with link target
    if not link_target:
        name_string = name
    else:
        name_string = "{} -> {}".format(name, link_target)
    #
    return "{}  {}  {} {}  {}  {}  {}".format(
        mode_string, nlink, user, group, size, datetime_string, name_string
    )



#! /usr/bin/env python
# Copyright (C) 2003-2022, Stefan Schwarzer <sschwarzer@sschwarzer.net>
# See the file LICENSE for licensing terms.

"""
setup.py - installation script for Python distutils
"""

import os
import sys

from distutils import core


_name = "ftputil"
_package = "ftputil"
_version = open("VERSION").read().strip()


core.setup(
    # Installation data
    name=_name,
    version=_version,
    packages=[_package],
    package_dir={_package: _package},
    # Metadata
    author="Stefan Schwarzer",
    author_email="sschwarzer@sschwarzer.net",
    url="https://ftputil.sschwarzer.net/",
    description="High-level FTP client library (virtual file system and more)",
    keywords="FTP, client, library, virtual file system",
    license="Open source (revised BSD license)",
    platforms=["Pure Python"],
    # See https://packaging.python.org/guides/distributing-packages-using-setuptools/#python-requires
    python_requires=">=3.6",
    long_description="""\
ftputil is a high-level FTP client library for the Python programming
language. ftputil implements a virtual file system for accessing FTP servers,
that is, it can generate file-like objects for remote files. The library
supports many functions similar to those in the os, os.path and
shutil modules. ftputil has convenience functions for conditional uploads
and downloads, and handles FTP clients and servers in different timezones.""",
    download_url="http://ftputil.sschwarzer.net/trac/attachment/wiki/Download/"
    "{}-{}.tar.gz?format=raw".format(_name, _version),
    classifiers=[
        # Commented-out for beta release
        "Development Status :: 5 - Production/Stable",
        #"Development Status :: 4 - Beta",
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Internet :: File Transfer Protocol (FTP)",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: System :: Filesystems",
    ],
)



# lrucache.py -- a simple LRU (Least-Recently-Used) cache class

# Copyright 2004 Evan Prodromou <evan@bad.dynu.ca>
#
# Copyright 2009-2018 Stefan Schwarzer <sschwarzer@sschwarzer.net>
# (some changes to the original version)

# Licensed under the Academic Free License 2.1

# Licensed for ftputil under the revised BSD license
# with permission by the author, Evan Prodromou. Many
# thanks, Evan! :-)
#
# The original file is available at
# http://pypi.python.org/pypi/lrucache/0.2 .

# arch-tag: LRU cache main module

"""a simple LRU (Least-Recently-Used) cache module

This module provides very simple LRU (Least-Recently-Used) cache
functionality.

An *in-memory cache* is useful for storing the results of an
'expensive' process (one that takes a lot of time or resources) for
later re-use. Typical examples are accessing data from the filesystem,
a database, or a network location. If you know you'll need to re-read
the data again, it can help to keep it in a cache.

You *can* use a Python dictionary as a cache for some purposes.
However, if the results you're caching are large, or you have a lot of
possible results, this can be impractical memory-wise.

An *LRU cache*, on the other hand, only keeps _some_ of the results in
memory, which keeps you from overusing resources. The cache is bounded
by a maximum size; if you try to add more values to the cache, it will
automatically discard the values that you haven't read or written to
in the longest time. In other words, the least-recently-used items are
discarded. [1]_

.. [1]: 'Discarded' here means 'removed from the cache'.

"""

import time


# The suffix after the hyphen denotes modifications by the
# ftputil project with respect to the original version.
__version__ = "0.2-15"
__all__ = ["CacheKeyError", "LRUCache", "DEFAULT_SIZE"]
__docformat__ = "reStructuredText en"

# Default size of a new LRUCache object, if no 'size' argument is given.
DEFAULT_SIZE = 16


class CacheKeyError(KeyError):
    """Error raised when cache requests fail.

    When a cache record is accessed which no longer exists (or never did),
    this error is raised. To avoid it, you may want to check for the existence
    of a cache record before reading or deleting it.
    """

    pass


class LRUCache:
    """Least-Recently-Used (LRU) cache.

    Instances of this class provide a least-recently-used (LRU) cache. They
    emulate a Python mapping type. You can use an LRU cache more or less like
    a Python dictionary, with the exception that objects you put into the
    cache may be discarded before you take them out.

    Some example usage::

    cache = LRUCache(32) # new cache
    cache['foo'] = get_file_contents('foo') # or whatever

    if 'foo' in cache: # if it's still in cache...
        # use cached version
        contents = cache['foo']
    else:
        # recalculate
        contents = get_file_contents('foo')
        # store in cache for next time
        cache['foo'] = contents

    print(cache.size) # Maximum size

    print(len(cache)) # 0 <= len(cache) <= cache.size

    cache.size = 10 # Auto-shrink on size assignment

    for i in range(50): # note: larger than cache size
        cache[i] = i

    if 0 not in cache: print('Zero was discarded.')

    if 42 in cache:
        del cache[42] # Manual deletion

    for j in cache:   # iterate (in LRU order)
        print(j, cache[j]) # iterator produces keys, not values
    """

    class _Node:
        """Record of a cached value. Not for public consumption."""

        def __init__(self, key, obj, timestamp, sort_key):
            object.__init__(self)
            self.key = key
            self.obj = obj
            self.atime = timestamp
            self.mtime = self.atime
            self._sort_key = sort_key

        def __lt__(self, other):
            # Seems to be preferred over `__cmp__`, at least in newer
            # Python versions. Uses only around 60 % of the time
            # with respect to `__cmp__`.
            # pylint: disable=protected-access
            return self._sort_key < other._sort_key

        def __repr__(self):
            return "<%s %s => %s (%s)>" % (
                self.__class__,
                self.key,
                self.obj,
                time.asctime(time.localtime(self.atime)),
            )

    def __init__(self, size=DEFAULT_SIZE):
        """Init the `LRUCache` object. `size` is the initial
        _maximum_ size of the cache. The size can be changed by
        setting the `size` attribute.
        """
        self.clear()
        # Maximum size of the cache. If more than 'size' elements are
        # added to the cache, the least-recently-used ones will be
        # discarded. This assignment implicitly checks the size value.
        self.size = size

    def clear(self):
        """Clear the cache, removing all elements.

        The `size` attribute of the cache isn't modified.
        """
        # pylint: disable=attribute-defined-outside-init
        self.__heap = []
        self.__dict = {}
        self.__counter = 0

    def _sort_key(self):
        """Return a new integer value upon every call.

        Cache nodes need a monotonically increasing time indicator.
        `time.time()` and `time.clock()` don't guarantee this in a
        platform-independent way.

        See http://ftputil.sschwarzer.net/trac/ticket/32 for details.
        """
        self.__counter += 1
        return self.__counter

    def __len__(self):
        """Return _current_ number of cache entries.

        This may be different from the value of the `size`
        attribute.
        """
        return len(self.__heap)

    def __contains__(self, key):
        """Return `True` if the item denoted by `key` is in the cache."""
        return key in self.__dict

    def __setitem__(self, key, obj):
        """Store item `obj` in the cache under the key `key`.

        If the number of elements after the addition of a new key
        would exceed the maximum cache size, the least recently
        used item in the cache is "forgotten".
        """
        heap = self.__heap
        dict_ = self.__dict
        if key in dict_:
            node = dict_[key]
            # Update node object in-place.
            node.obj = obj
            node.atime = time.time()
            node.mtime = node.atime
            # pylint: disable=protected-access
            node._sort_key = self._sort_key()
        else:
            # The size of the heap can be at most the value of
            # `self.size` because `__setattr__` decreases the cache
            # size if the new size value is smaller; so we don't
            # need a loop _here_.
            if len(heap) == self.size:
                lru_node = min(heap)
                heap.remove(lru_node)
                del dict_[lru_node.key]
            node = self._Node(key, obj, time.time(), self._sort_key())
            dict_[key] = node
            heap.append(node)

    def __getitem__(self, key):
        """Return the item stored under `key` key.

        If no such key is present in the cache, raise a
        `CacheKeyError`.
        """
        if not key in self.__dict:
            raise CacheKeyError(key)
        else:
            node = self.__dict[key]
            # Update node object in-place.
            node.atime = time.time()
            # pylint: disable=protected-access
            node._sort_key = self._sort_key()
            return node.obj

    def __delitem__(self, key):
        """Delete the item stored under `key` key.

        If no such key is present in the cache, raise a
        `CacheKeyError`.
        """
        if not key in self.__dict:
            raise CacheKeyError(key)
        else:
            node = self.__dict[key]
            self.__heap.remove(node)
            del self.__dict[key]
            return node.obj

    def __iter__(self):
        """Iterate over the cache, from the least to the most
        recently accessed item.
        """
        self.__heap.sort()
        for node in self.__heap:
            yield node.key

    def __setattr__(self, name, value):
        """If the name of the attribute is "size", set the
        _maximum_ size of the cache to the supplied value.
        """
        object.__setattr__(self, name, value)
        # Automagically shrink heap on resize.
        if name == "size":
            size = value
            if not isinstance(size, int):
                raise TypeError("cache size (%r) must be an integer" % size)
            if size <= 0:
                raise ValueError("cache size (%d) must be positive" % size)
            heap = self.__heap
            dict_ = self.__dict
            # Do we need to remove anything at all?
            if len(heap) <= self.size:
                return
            # Remove enough nodes to reach the new size.
            heap.sort()
            node_count_to_remove = len(heap) - self.size
            for node in heap[:node_count_to_remove]:
                del dict_[node.key]
            del heap[:node_count_to_remove]

    def __repr__(self):
        return "<%s (%d elements)>" % (str(self.__class__), len(self.__heap))

    def mtime(self, key):
        """Return the last modification time for the cache record with key.

        May be useful for cache instances where the stored values can get
        "stale", such as caching file or network resource contents.
        """
        if not key in self.__dict:
            raise CacheKeyError(key)
        else:
            node = self.__dict[key]
            return node.mtime


if __name__ == "__main__":
    cache = LRUCache(25)
    print(cache)
    for i in range(50):
        cache[i] = str(i)
    print(cache)
    if 46 in cache:
        del cache[46]
    print(cache)
    cache.size = 10
    print(cache)
    cache[46] = "46"
    print(cache)
    print(len(cache))
    for c in cache:
        print(c)
    print(cache)
    print(cache.mtime(46))
    for c in cache:
        print(c)



# Copyright (C) 2013-2021, Stefan Schwarzer
# and ftputil contributors (see `doc/contributors.txt`)
# See the file LICENSE for licensing terms.

import os

import pytest

import ftputil.path_encoding
import ftputil.tool


DEFAULT_ENCODING = ftputil.path_encoding.DEFAULT_ENCODING


class Path(os.PathLike):
    """
    Helper class for `TestSameStringTypeAs`
    """

    def __init__(self, type_source):
        self._type_source = type_source

    def __fspath__(self):
        return self._type_source


same_string_type_as = ftputil.tool.same_string_type_as


class TestSameStringTypeAs:
    @staticmethod
    def _test_string(type_source, string, expected_result):
        """
        Check if the result from `tool.same_string_type_as(type_source,
        string)` is the same as `expected_result`.

        `type_source` must be a `bytes` or `str` object.
        """
        result = ftputil.tool.same_string_type_as(type_source, path)
        assert result == expected_result

    def test_to_bytes(self):
        assert same_string_type_as(b"abc", "def", encoding=DEFAULT_ENCODING) == b"def"

    def test_to_str(self):
        assert same_string_type_as("abc", b"def", encoding=DEFAULT_ENCODING) == "def"

    def test_both_bytes_type(self):
        assert same_string_type_as(b"abc", b"def", encoding=DEFAULT_ENCODING) == b"def"

    def test_both_str_type(self):
        assert same_string_type_as("abc", "def", encoding=DEFAULT_ENCODING) == "def"


as_str = ftputil.tool.as_str
as_str_path = ftputil.tool.as_str_path


class TestAsStr:
    def test_from_bytes(self):
        assert as_str(b"abc", encoding=DEFAULT_ENCODING) == "abc"
        assert as_str_path(b"abc", encoding=DEFAULT_ENCODING) == "abc"

    def test_from_str(self):
        assert as_str("abc", encoding=DEFAULT_ENCODING) == "abc"
        assert as_str_path("abc", encoding=DEFAULT_ENCODING) == "abc"

    def test_from_bytes_path(self):
        assert as_str_path(Path(b"abc"), encoding=DEFAULT_ENCODING) == "abc"

    def test_from_str_path(self):
        assert as_str_path(Path("abc"), encoding=DEFAULT_ENCODING) == "abc"

    def test_type_error(self):
        with pytest.raises(TypeError):
            as_str(1, encoding=DEFAULT_ENCODING)



# Copyright (C) 2007-2018, Stefan Schwarzer <sschwarzer@sschwarzer.net>
# and ftputil contributors (see `doc/contributors.txt`)
# See the file LICENSE for licensing terms.

"""
Tools for syncing combinations of local and remote directories.

*** WARNING: This is an unfinished in-development version!
"""

# Sync combinations:
# - remote -> local (download)
# - local -> remote (upload)
# - remote -> remote
# - local -> local (maybe implicitly possible due to design, but not targeted)

import os
import shutil

from ftputil import FTPHost
import ftputil.error

__all__ = ["FTPHost", "LocalHost", "Syncer"]


# Used for copying file objects; value is 64 KB.
CHUNK_SIZE = 64 * 1024


class LocalHost:
    """
    Provide an API for local directories and files so we can use the same code
    as for `FTPHost` instances.
    """

    def open(self, path, mode):
        """
        Return a Python file object for file name `path`, opened in mode
        `mode`.
        """
        # This is the built-in `open` function, not `os.open`!
        return open(path, mode)

    def time_shift(self):
        """
        Return the time shift value (see methods `set_time_shift` and
        `time_shift` in class `FTPHost` for a definition). By definition, the
        value is zero for local file systems.
        """
        return 0.0

    def __getattr__(self, attr):
        return getattr(os, attr)


class Syncer:
    """
    Control synchronization between combinations of local and remote
    directories and files.
    """

    def __init__(self, source, target):
        """
        Init the `FTPSyncer` instance.

        Each of `source` and `target` is either an `FTPHost` or a `LocalHost`
        object. The source and target directories, resp. have to be set with
        the `chdir` command before passing them in. The semantics is so that
        the items under the source directory will show up under the target
        directory after the synchronization (unless there's an error).
        """
        self._source = source
        self._target = target

    def _mkdir(self, target_dir):
        """
        Try to create the target directory `target_dir`. If it already exists,
        don't do anything. If the directory is present but it's actually a
        file, raise a `SyncError`.
        """
        # TODO: Handle setting of target mtime according to source mtime
        # (beware of rootdir anomalies; try to handle them as well).
        if self._target.path.isfile(target_dir):
            raise ftputil.error.SyncError(
                "target dir '{}' is actually a file".format(target_dir)
            )
        # Deliberately use an `isdir` test instead of `try/except`. The latter
        # approach might mask other errors we want to see, e. g. insufficient
        # permissions.
        if not self._target.path.isdir(target_dir):
            self._target.mkdir(target_dir)

    def _sync_file(self, source_file, target_file):
        # XXX: This duplicates code from `FTPHost._copyfileobj`. Maybe
        # implement the upload and download methods in terms of `_sync_file`,
        # or maybe not?
        # TODO: Handle `IOError`s
        # TODO: Handle conditional copy
        # TODO: Handle setting of target mtime according to source mtime
        # (beware of rootdir anomalies; try to handle them as well).
        source = self._source.open(source_file, "rb")
        try:
            target = self._target.open(target_file, "wb")
            try:
                shutil.copyfileobj(source, target, length=CHUNK_SIZE)
            finally:
                target.close()
        finally:
            source.close()

    def _fix_sep_for_target(self, path):
        """
        Return the string `path` with appropriate path separators for the
        target file system.
        """
        return path.replace(self._source.sep, self._target.sep)

    def _sync_tree(self, source_dir, target_dir):
        """
        Synchronize the source and the target directory tree by updating the
        target to match the source as far as possible.

        Current limitations:
        - _don't_ delete items which are on the target path but not on the
          source path
        - files are always copied, the modification timestamps are not compared
        - all files are copied in binary mode, never in ASCII/text mode
        - incomplete error handling
        """
        self._mkdir(target_dir)
        for dirpath, dir_names, file_names in self._source.walk(source_dir):
            for dir_name in dir_names:
                inner_source_dir = self._source.path.join(dirpath, dir_name)
                inner_target_dir = inner_source_dir.replace(source_dir, target_dir, 1)
                inner_target_dir = self._fix_sep_for_target(inner_target_dir)
                self._mkdir(inner_target_dir)
            for file_name in file_names:
                source_file = self._source.path.join(dirpath, file_name)
                target_file = source_file.replace(source_dir, target_dir, 1)
                target_file = self._fix_sep_for_target(target_file)
                self._sync_file(source_file, target_file)

    def sync(self, source_path, target_path):
        """
        Synchronize `source_path` and `target_path` (both are strings, each
        denoting a directory or file path), i. e. update the target path so
        that it's a copy of the source path.

        This method handles both directory trees and single files.
        """
        # TODO: Handle making of missing intermediate directories.
        source_path = self._source.path.abspath(source_path)
        target_path = self._target.path.abspath(target_path)
        if self._source.path.isfile(source_path):
            self._sync_file(source_path, target_path)
        else:
            self._sync_tree(source_path, target_path)



# Copyright (C) 2002-2020, Stefan Schwarzer <sschwarzer@sschwarzer.net>
# and ftputil contributors (see `doc/contributors.txt`)
# See the file LICENSE for licensing terms.

"""
ftputil.stat - stat result, parsers, and FTP stat'ing for `ftputil`
"""

import datetime
import math
import re
import stat

import ftputil.error
import ftputil.stat_cache


# These can be used to write custom parsers.
__all__ = ["StatResult", "Parser", "UnixParser", "MSParser"]


# Datetime precision values in seconds.
MINUTE_PRECISION = 60
DAY_PRECISION = 24 * 60 * 60
UNKNOWN_PRECISION = None


class StatResult(tuple):
    """
    Support class resembling a tuple like that returned from `os.(l)stat`.
    """

    _index_mapping = {
        "st_mode": 0,
        "st_ino": 1,
        "st_dev": 2,
        "st_nlink": 3,
        "st_uid": 4,
        "st_gid": 5,
        "st_size": 6,
        "st_atime": 7,
        "st_mtime": 8,
        "st_ctime": 9,
        "_st_name": 10,
        "_st_target": 11,
    }

    def __init__(self, sequence):
        # Don't call `__init__` via `super`. Construction from a sequence is
        # implicitly handled by `tuple.__new__`, not `tuple.__init__`.
        # pylint: disable=super-init-not-called
        #
        # Use `sequence` parameter to remain compatible to `__new__` interface.
        # pylint: disable=unused-argument
        #
        # These may be overwritten in a `Parser.parse_line` method.
        self._st_name = ""
        self._st_target = None
        self._st_mtime_precision = UNKNOWN_PRECISION

    def __getattr__(self, attr_name):
        if attr_name in self._index_mapping:
            return self[self._index_mapping[attr_name]]
        else:
            raise AttributeError(
                "'StatResult' object has no attribute '{}'".format(attr_name)
            )

    def __repr__(self):
        # "Invert" `_index_mapping` so that we can look up the names for the
        # tuple indices.
        index_to_name = dict((v, k) for k, v in self._index_mapping.items())
        argument_strings = []
        for index, item in enumerate(self):
            argument_strings.append("{}={!r}".format(index_to_name[index], item))
        return "{}({})".format(type(self).__name__, ", ".join(argument_strings))


#
# FTP directory parsers
#
class Parser:
    """
    Represent a parser for directory lines. Parsers for specific directory
    formats inherit from this class.
    """

    # Map month abbreviations to month numbers.
    _month_numbers = {
        "jan": 1,
        "feb": 2,
        "mar": 3,
        "apr": 4,
        "may": 5,
        "jun": 6,
        "jul": 7,
        "aug": 8,
        "sep": 9,
        "oct": 10,
        "nov": 11,
        "dec": 12,
    }

    _total_regex = re.compile(r"^total\s+\d+")

    def ignores_line(self, line):
        """
        Return a true value if the line should be ignored, i. e. is assumed to
        _not_ contain actual directory/file/link data. A typical example are
        summary lines like "total 23" which are emitted by some FTP servers.

        If the line should be used to extract stat data from it, return a false
        value.
        """
        # Ignore empty lines stemming from only a line break.
        if not line.strip():
            # Yes, ignore the line if it's empty.
            return True
        # Either a `_SRE_Match` instance or `None`
        match = self._total_regex.search(line)
        return bool(match)

    def parse_line(self, line, time_shift=0.0):
        """
        Return a `StatResult` object as derived from the string `line`. The
        parser code to use depends on the directory format the FTP server
        delivers (also see examples at end of file).

        If the given text line can't be parsed, raise a `ParserError`.

        For the definition of `time_shift` see the docstring of
        `FTPHost.set_time_shift` in `ftputil.py`. Not all parsers use the
        `time_shift` parameter.
        """
        raise NotImplementedError("must be defined by subclass")

    #
    # Helper methods for parts of a directory listing line
    #
    def parse_unix_mode(self, mode_string):
        """
        Return an integer from the `mode_string`, compatible with the `st_mode`
        value in stat results. Such a mode string may look like "drwxr-xr-x".

        If the mode string can't be parsed, raise an
        `ftputil.error.ParserError`.
        """
        # Allow derived classes to make use of `self`.
        # pylint: disable=no-self-use
        if len(mode_string) != 10:
            raise ftputil.error.ParserError(
                "invalid mode string '{}'".format(mode_string)
            )
        st_mode = 0
        # TODO: Add support for "S" and sticky bit ("t", "T").
        for bit in mode_string[1:10]:
            bit = bit != "-"
            st_mode = (st_mode << 1) + bit
        if mode_string[3] == "s":
            st_mode = st_mode | stat.S_ISUID
        if mode_string[6] == "s":
            st_mode = st_mode | stat.S_ISGID
        file_type_to_mode = {
            "b": stat.S_IFBLK,
            "c": stat.S_IFCHR,
            "d": stat.S_IFDIR,
            "l": stat.S_IFLNK,
            "p": stat.S_IFIFO,
            "s": stat.S_IFSOCK,
            "-": stat.S_IFREG,
            # Ignore types which `ls` can't make sense of (assuming the FTP
            # server returns listings like `ls` does).
            "?": 0,
        }
        file_type = mode_string[0]
        if file_type in file_type_to_mode:
            st_mode = st_mode | file_type_to_mode[file_type]
        else:
            raise ftputil.error.ParserError(
                "unknown file type character '{}'".format(file_type)
            )
        return st_mode

    # pylint: disable=no-self-use
    def _as_int(self, int_string, int_description):
        """
        Return `int_string` converted to an integer.

        If it can't be converted, raise a `ParserError`, using
        `int_description` in the error message. For example, if the integer
        value is a day, pass "day" for `int_description`.
        """
        try:
            return int(int_string)
        except ValueError:
            raise ftputil.error.ParserError(
                "non-integer {} value {!r}".format(int_description, int_string)
            )

    @staticmethod
    def _datetime(year, month, day, hour, minute, second):
        """
        Return UTC `datetime.datetime` object for the given year, month, day,
        hour, minute and second.

        If there are invalid values, for example minute > 59, raise a
        `ParserError`.
        """
        try:
            return datetime.datetime(
                year, month, day, hour, minute, second, tzinfo=datetime.timezone.utc
            )
        except ValueError:
            invalid_datetime = (
                f"{year:04d}-{month:02d}-{day:02d} "
                f"{hour:02d}:{minute:02d}:{second:02d}"
            )
            raise ftputil.error.ParserError(
                "invalid datetime {0!r}".format(invalid_datetime)
            )

    def parse_unix_time(
        self, month_abbreviation, day, year_or_time, time_shift, with_precision=False
    ):
        """
        Return a floating point number, like from `time.mktime`, by parsing the
        string arguments `month_abbreviation`, `day` and `year_or_time`. The
        parameter `time_shift` is the difference "time on server" - "time on
        client" and is available as the `time_shift` parameter in the
        `parse_line` interface.

        If `with_precision` is true return a two-element tuple consisting of
        the floating point number as described in the previous paragraph and
        the precision of the time in seconds. The default is `False` for
        backward compatibility with custom parsers.

        The precision value takes into account that, for example, a time string
        like "May 26  2005" has only a precision of one day. This information
        is important for the `upload_if_newer` and `download_if_newer` methods
        in the `FTPHost` class.

        Times in Unix-style directory listings typically have one of these
        formats:

        - "Nov 23 02:33" (month name, day of month, time)

        - "May 26  2005" (month name, day of month, year)

        If this method can't make sense of the given arguments, it raises an
        `ftputil.error.ParserError`.
        """
        try:
            month = self._month_numbers[month_abbreviation.lower()]
        except KeyError:
            raise ftputil.error.ParserError(
                "invalid month abbreviation {0!r}".format(month_abbreviation)
            )
        day = self._as_int(day, "day")
        year_is_known = ":" not in year_or_time
        if year_is_known:
            # `year_or_time` is really a year.
            st_mtime_precision = DAY_PRECISION
            server_year, hour, minute = self._as_int(year_or_time, "year"), 0, 0
        else:
            # `year_or_time` is a time hh:mm.
            st_mtime_precision = MINUTE_PRECISION
            hour, minute = year_or_time.split(":")
            year, hour, minute = (
                None,
                self._as_int(hour, "hour"),
                self._as_int(minute, "minute"),
            )
            # First assume the year of the directory/file is the current year.
            server_now = datetime.datetime.utcnow().replace(
                tzinfo=datetime.timezone.utc
            ) + datetime.timedelta(seconds=time_shift)
            server_year = server_now.year
            # If the server datetime derived from this year seems to be in the
            # future, subtract one year.
            #
            # Things to consider:
            #
            # Since the server time will be rounded down to full minutes, apply
            # the same truncation to the presumed current server time.
            #
            # Due to possible small errors in the time setting of the server
            # (not only because of the parsing), there will always be a small
            # time window in which we can't be sure whether the parsed time is
            # in the future or not. Make a mistake of one second and the time
            # is off one year.
            #
            # To resolve this conflict, if in doubt assume that a directory or
            # file has just been created, not that it is by chance to the
            # minute one year old.
            #
            # Hence, add a small time difference that must be exceeded in order
            # to assume the time is in the future. This time difference is
            # arbitrary, but we have to assume _some_ value.
            if self._datetime(
                server_year, month, day, hour, minute, 0
            ) > server_now.replace(second=0) + datetime.timedelta(seconds=120):
                server_year -= 1
        # The time shift is the time difference the server is ahead. So to get
        # back to client time (UTC), subtract the time shift. The calculation
        # is supposed to be the same for negative time shifts; in this case we
        # subtract a negative time shift, i. e. add the absolute value of the
        # time shift to the server date time.
        server_utc_datetime = self._datetime(
            server_year, month, day, hour, minute, 0
        ) - datetime.timedelta(seconds=time_shift)
        st_mtime = server_utc_datetime.timestamp()
        # If we had a datetime before the epoch, the resulting value 0.0
        # doesn't tell us anything about the precision.
        if st_mtime < 0.0:
            st_mtime_precision = UNKNOWN_PRECISION
            st_mtime = 0.0
        #
        if with_precision:
            return st_mtime, st_mtime_precision
        else:
            return st_mtime

    def parse_ms_time(self, date, time_, time_shift, with_precision=False):
        """
        Return a floating point number, like from `time.mktime`, by parsing the
        string arguments `date` and `time_`. The parameter `time_shift` is the
        difference

            "time on server" - "time on client"

        and can be set as the `time_shift` parameter in the `parse_line`
        interface.

        If `with_precision` is true return a two-element tuple consisting of
        the floating point number as described in the previous paragraph and
        the precision of the time in seconds. The default is `False` for
        backward compatibility with custom parsers.

        The precision value takes into account that, for example, a time string
        like "10-23-2001 03:25PM" has only a precision of one minute. This
        information is important for the `upload_if_newer` and
        `download_if_newer` methods in the `FTPHost` class.

        Usually, the returned precision is `MINUTE_PRECISION`, except when the
        date is before the epoch, in which case the returned `st_mtime` value
        is set to 0.0 and the precision to `UNKNOWN_PRECISION`.

        Times in MS-style directory listings typically have the format
        "10-23-01 03:25PM" (month-day_of_month-two_digit_year, hour:minute,
        am/pm).

        If this method can't make sense of the given arguments, it raises an
        `ftputil.error.ParserError`.
        """
        # Derived classes might want to use `self`.
        # pylint: disable=no-self-use
        #
        # Derived classes may need access to `time_shift`.
        # pylint: disable=unused-argument
        #
        # For the time being, I don't add a `with_precision` parameter as in
        # the MS parser because the precision for the DOS format is always a
        # minute and can be set in `MSParser.parse_line`. Should you find
        # yourself needing support for `with_precision` for a derived class,
        # please send a mail (see ftputil.txt/html).
        month, day, year = [
            self._as_int(part, "year/month/day") for part in date.split("-")
        ]
        if year >= 1000:
            # We have a four-digit year, so no need for heuristics.
            pass
        elif year >= 70:
            year = 1900 + year
        else:
            year = 2000 + year
        try:
            hour, minute, am_pm = time_[0:2], time_[3:5], time_[5]
        except IndexError:
            raise ftputil.error.ParserError("invalid time string '{}'".format(time_))
        hour, minute = (self._as_int(hour, "hour"), self._as_int(minute, "minute"))
        if hour == 12 and am_pm == "A":
            hour = 0
        if hour != 12 and am_pm == "P":
            hour += 12
        server_datetime = self._datetime(year, month, day, hour, minute, 0)
        client_datetime = server_datetime - datetime.timedelta(seconds=time_shift)
        st_mtime = client_datetime.timestamp()
        if st_mtime < 0.0:
            st_mtime_precision = UNKNOWN_PRECISION
            st_mtime = 0.0
        else:
            st_mtime_precision = MINUTE_PRECISION
        if with_precision:
            return st_mtime, st_mtime_precision
        else:
            return st_mtime


class UnixParser(Parser):
    """
    `Parser` class for Unix-specific directory format.
    """

    @staticmethod
    def _split_line(line):
        """
        Split a line in metadata, nlink, user, group, size, month, day,
        year_or_time and name and return the result as an nine-element list of
        these values. If the name is a link, it will be encoded as a string
        "link_name -> link_target".
        """
        # This method encapsulates the recognition of an unusual Unix format
        # variant (see ticket http://ftputil.sschwarzer.net/trac/ticket/12 ).
        line_parts = line.split()
        FIELD_COUNT_WITHOUT_USERID = 8
        FIELD_COUNT_WITH_USERID = FIELD_COUNT_WITHOUT_USERID + 1
        if len(line_parts) < FIELD_COUNT_WITHOUT_USERID:
            # No known Unix-style format
            raise ftputil.error.ParserError("line '{}' can't be parsed".format(line))
        # If we have a valid format (either with or without user id field), the
        # field with index 5 is either the month abbreviation or a day.
        try:
            int(line_parts[5])
        except ValueError:
            # Month abbreviation, "invalid literal for int"
            line_parts = line.split(None, FIELD_COUNT_WITH_USERID - 1)
        else:
            # Day
            line_parts = line.split(None, FIELD_COUNT_WITHOUT_USERID - 1)
            USER_FIELD_INDEX = 2
            line_parts.insert(USER_FIELD_INDEX, None)
        return line_parts

    def parse_line(self, line, time_shift=0.0):
        """
        Return a `StatResult` instance corresponding to the given text line.
        The `time_shift` value is needed to determine to which year a datetime
        without an explicit year belongs.

        If the line can't be parsed, raise a `ParserError`.
        """
        # The local variables are rather simple.
        # pylint: disable=too-many-locals
        try:
            (
                mode_string,
                nlink,
                user,
                group,
                size,
                month,
                day,
                year_or_time,
                name,
            ) = self._split_line(line)
        # We can get a `ValueError` here if the name is blank (see ticket #69).
        # This is a strange use case, but at least we should raise the
        # exception the docstring mentions.
        except ValueError as exc:
            raise ftputil.error.ParserError(str(exc))
        # st_mode
        st_mode = self.parse_unix_mode(mode_string)
        # st_ino, st_dev, st_nlink, st_uid, st_gid, st_size, st_atime
        st_ino = None
        st_dev = None
        st_nlink = int(nlink)
        st_uid = user
        st_gid = group
        st_size = int(size)
        st_atime = None
        # st_mtime
        st_mtime, st_mtime_precision = self.parse_unix_time(
            month, day, year_or_time, time_shift, with_precision=True
        )
        # st_ctime
        st_ctime = None
        # st_name
        if name.count(" -> ") > 1:
            # If we have more than one arrow we can't tell where the link name
            # ends and the target name starts.
            raise ftputil.error.ParserError(
                '''name '{}' contains more than one "->"'''.format(name)
            )
        elif name.count(" -> ") == 1:
            st_name, st_target = name.split(" -> ")
        else:
            st_name, st_target = name, None
        stat_result = StatResult(
            (
                st_mode,
                st_ino,
                st_dev,
                st_nlink,
                st_uid,
                st_gid,
                st_size,
                st_atime,
                st_mtime,
                st_ctime,
            )
        )
        # These attributes are kind of "half-official". I'm not sure whether
        # they should be used by ftputil client code.
        # pylint: disable=protected-access
        stat_result._st_mtime_precision = st_mtime_precision
        stat_result._st_name = st_name
        stat_result._st_target = st_target
        return stat_result


class MSParser(Parser):
    """
    `Parser` class for MS-specific directory format.
    """

    def parse_line(self, line, time_shift=0.0):
        """
        Return a `StatResult` instance corresponding to the given text line
        from a FTP server which emits "Microsoft format" (see end of file).

        If the line can't be parsed, raise a `ParserError`.

        The parameter `time_shift` isn't used in this method but is listed for
        compatibility with the base class.
        """
        # The local variables are rather simple.
        # pylint: disable=too-many-locals
        try:
            date, time_, dir_or_size, name = line.split(None, 3)
        except ValueError:
            # "unpack list of wrong size"
            raise ftputil.error.ParserError("line '{}' can't be parsed".format(line))
        # st_mode
        #  Default to read access only; in fact, we can't tell.
        st_mode = 0o400
        if dir_or_size == "<DIR>":
            st_mode = st_mode | stat.S_IFDIR
        else:
            st_mode = st_mode | stat.S_IFREG
        # st_ino, st_dev, st_nlink, st_uid, st_gid
        st_ino = None
        st_dev = None
        st_nlink = None
        st_uid = None
        st_gid = None
        # st_size
        if dir_or_size != "<DIR>":
            try:
                st_size = int(dir_or_size)
            except ValueError:
                raise ftputil.error.ParserError("invalid size {}".format(dir_or_size))
        else:
            st_size = None
        # st_atime
        st_atime = None
        # st_mtime
        st_mtime, st_mtime_precision = self.parse_ms_time(
            date, time_, time_shift, with_precision=True
        )
        # st_ctime
        st_ctime = None
        stat_result = StatResult(
            (
                st_mode,
                st_ino,
                st_dev,
                st_nlink,
                st_uid,
                st_gid,
                st_size,
                st_atime,
                st_mtime,
                st_ctime,
            )
        )
        # These attributes are kind of "half-official". I'm not sure whether
        # they should be used by ftputil client code.
        # pylint: disable=protected-access
        # _st_name and _st_target
        stat_result._st_name = name
        stat_result._st_target = None
        # mtime precision in seconds
        stat_result._st_mtime_precision = st_mtime_precision
        return stat_result


#
# Stat'ing operations for files on an FTP server
#
class _Stat:
    """
    Methods for stat'ing directories, links and regular files.
    """

    # pylint: disable=protected-access

    def __init__(self, host):
        self._host = host
        self._path = host.path
        # Use the Unix directory parser by default.
        self._parser = UnixParser()
        # Allow one chance to switch to another parser if the default doesn't
        # work.
        self._allow_parser_switching = True
        # Cache only lstat results. `stat` works locally on `lstat` results.
        self._lstat_cache = ftputil.stat_cache.StatCache()

    def _host_dir(self, path):
        """
        Return a list of lines, as fetched by FTP's `LIST` command, when
        applied to `path`.
        """
        return self._host._dir(path)

    def _stat_results_from_dir(self, path):
        """
        Yield stat results extracted from the directory listing `path`. Omit
        the special entries for the directory itself and its parent directory.
        """
        lines = self._host_dir(path)
        # `cache` is the "high-level" `StatCache` object whereas `cache._cache`
        # is the "low-level" `LRUCache` object.
        cache = self._lstat_cache
        # Auto-grow cache if the cache up to now can't hold as many entries as
        # there are in the directory `path`.
        if cache._enabled and len(lines) >= cache._cache.size:
            new_size = int(math.ceil(1.1 * len(lines)))
            cache.resize(new_size)
        # Yield stat results from lines.
        for line in lines:
            if self._parser.ignores_line(line):
                continue
            # Although for a `listdir` call we're only interested in the names,
            # use the `time_shift` parameter to store the correct timestamp
            # values in the cache.
            stat_result = self._parser.parse_line(line, self._host.time_shift())
            # Skip entries "." and "..".
            if stat_result._st_name in [self._host.curdir, self._host.pardir]:
                continue
            loop_path = self._path.join(path, stat_result._st_name)
            # No-op if cache is disabled.
            cache[loop_path] = stat_result
            yield stat_result

    # The methods `listdir`, `lstat` and `stat` come in two variants. The
    # methods `_real_listdir`, `_real_lstat` and `_real_stat` use the currently
    # set parser to get the directories/files of the requested directory, the
    # lstat result or the stat result, respectively.
    #
    # Additionally, we have the methods `_listdir`, `_lstat` and `_stat`, which
    # wrap the above `_real_*` methods. _For example_, `_listdir` calls
    # `_real_listdir`. If `_real_listdir` can't parse the directory lines and a
    # parser hasn't been fixed yet, `_listdir` switches to the MS parser and
    # calls `_real_listdir` again.
    #
    # There are two important conditions to watch out for:
    #
    # - If the user explicitly set a different parser with
    #   `FTPHost.set_parser`, parser switching is disabled after that and
    #   `_listdir` etc. only call "their" method once with the fixed parser.
    #
    # - A `_real_*` call will fail if there's no directory line at all in the
    #   given directory. In that case, we can't tell whether the default parser
    #   was appropriate or not. Hence parser switching will still be allowed
    #   until we encounter a directory that has directories/files/links in it.

    def _real_listdir(self, path):
        """
        Return a list of directories, files etc. in the directory named `path`.

        Like `os.listdir` the returned list elements have the type of the path
        argument.

        If the directory listing from the server can't be parsed, raise a
        `ParserError`.
        """
        # We _can't_ put this check into `FTPHost._dir`; see its docstring.
        path = self._path.abspath(path)
        # `listdir` should only be allowed for directories and links to them.
        if not self._path.isdir(path):
            raise ftputil.error.PermanentError(
                "550 {}: no such directory or wrong directory parser used".format(path)
            )
        # Set up for `for` loop.
        names = []
        for stat_result in self._stat_results_from_dir(path):
            st_name = stat_result._st_name
            names.append(st_name)
        return names

    def _real_lstat(self, path, _exception_for_missing_path=True):
        """
        Return an object similar to that returned by `os.lstat`.

        If the directory listing from the server can't be parsed, raise a
        `ParserError`. If the directory can be parsed and the `path` is not
        found, raise a `PermanentError`. That means that if the directory
        containing `path` can't be parsed we get a `ParserError`, independent
        on the presence of `path` on the server.

        (`_exception_for_missing_path` is an implementation aid and _not_
        intended for use by ftputil clients.)
        """
        path = self._path.abspath(path)
        # If the path is in the cache, return the lstat result.
        if path in self._lstat_cache:
            return self._lstat_cache[path]
        # Note: (l)stat works by going one directory up and parsing the output
        # of an FTP `LIST` command. Unfortunately, it is not possible to do
        # this for the root directory `/`.
        if path == "/":
            raise ftputil.error.RootDirError("can't stat remote root directory")
        dirname, basename = self._path.split(path)
        # If even the directory doesn't exist and we don't want the exception,
        # treat it the same as if the path wasn't found in the directory's
        # contents (compare below). The use of `isdir` here causes a recursion
        # but that should be ok because that will at the latest stop when we've
        # gotten to the root directory.
        if not self._path.isdir(dirname) and not _exception_for_missing_path:
            return None
        # Loop through all lines of the directory listing. We probably won't
        # need all lines for the particular path but we want to collect as many
        # stat results in the cache as possible.
        lstat_result_for_path = None
        # FIXME: Here we try to list the contents of `dirname` even though the
        # above `isdir` call might/could have shown that the directory doesn't
        # exist. This may be related to ticket #108. That said, we may need to
        # consider virtual directories here (see tickets #86 / #87).
        for stat_result in self._stat_results_from_dir(dirname):
            # Needed to work without cache or with disabled cache.
            if stat_result._st_name == basename:
                lstat_result_for_path = stat_result
        if lstat_result_for_path is not None:
            return lstat_result_for_path
        # Path was not found during the loop.
        if _exception_for_missing_path:
            # TODO: Use FTP `LIST` command on the file to implicitly use the
            # usual status code of the server for missing files (450 vs. 550).
            raise ftputil.error.PermanentError(
                "550 {}: no such file or directory".format(path)
            )
        else:
            # Be explicit. Returning `None` is a signal for
            # `_Path.exists/isfile/isdir/islink` that the path was not found.
            # If we would raise an exception, there would be no distinction
            # between a missing path or a more severe error in the code above.
            return None

    def _real_stat(self, path, _exception_for_missing_path=True):
        """
        Return info from a "stat" call on `path`.

        If the directory containing `path` can't be parsed, raise a
        `ParserError`. If the listing can be parsed but the `path` can't be
        found, raise a `PermanentError`. Also raise a `PermanentError` if
        there's an endless (cyclic) chain of symbolic links "behind" the
        `path`.

        (`_exception_for_missing_path` is an implementation aid and _not_
        intended for use by ftputil clients.)
        """
        # Save for error message.
        original_path = path
        # Most code in this method is used to detect recursive link structures.
        visited_paths = set()
        while True:
            # Stat the link if it is one, else the file/directory.
            lstat_result = self._real_lstat(path, _exception_for_missing_path)
            if lstat_result is None:
                return None
            # If the file is not a link, the `stat` result is the same as the
            # `lstat` result.
            if not stat.S_ISLNK(lstat_result.st_mode):
                return lstat_result
            # If we stat'ed a link, calculate a normalized path for the file
            # the link points to.
            dirname, _ = self._path.split(path)
            path = self._path.join(dirname, lstat_result._st_target)
            path = self._path.abspath(self._path.normpath(path))
            # Check for cyclic structure.
            if path in visited_paths:
                # We had seen this path already.
                raise ftputil.error.RecursiveLinksError(
                    "recursive link structure detected for remote path '{}'".format(
                        original_path
                    )
                )
            # Remember the path we have encountered.
            visited_paths.add(path)

    def __call_with_parser_retry(self, method, *args, **kwargs):
        """
        Call `method` with the `args` and `kwargs` once. If that results in a
        `ParserError` and only one parser has been used yet, try the other
        parser. If that still fails, propagate the `ParserError`.
        """
        # Do _not_ set `_allow_parser_switching` in a `finally` clause! This
        # would cause a `PermanentError` due to a not-found file in an empty
        # directory to finally establish the parser - which is wrong.
        try:
            result = method(*args, **kwargs)
            # If a `listdir` call didn't find anything, we can't say anything
            # about the usefulness of the parser.
            if (method is not self._real_listdir) and result:
                self._allow_parser_switching = False
            return result
        except ftputil.error.ParserError:
            if self._allow_parser_switching:
                self._allow_parser_switching = False
                self._parser = MSParser()
                return method(*args, **kwargs)
            else:
                raise

    # Client code should never use these methods, but only the corresponding
    # methods without the leading underscore in the `FTPHost` class.

    def _listdir(self, path):
        """
        Return a list of items in `path`.

        Raise a `PermanentError` if the path doesn't exist, but maybe raise
        other exceptions depending on the state of the server (e. g. timeout).
        """
        return self.__call_with_parser_retry(self._real_listdir, path)

    def _lstat(self, path, _exception_for_missing_path=True):
        """
        Return a `StatResult` without following links.

        Raise a `PermanentError` if the path doesn't exist, but maybe raise
        other exceptions depending on the state of the server (e. g. timeout).
        """
        return self.__call_with_parser_retry(
            self._real_lstat, path, _exception_for_missing_path
        )

    def _stat(self, path, _exception_for_missing_path=True):
        """
        Return a `StatResult` with following links.

        Raise a `PermanentError` if the path doesn't exist, but maybe raise
        other exceptions depending on the state of the server (e. g. timeout).
        """
        return self.__call_with_parser_retry(
            self._real_stat, path, _exception_for_missing_path
        )
